﻿Option Strict On
Imports Microsoft.Kinect
Imports Microsoft.Kinect.Toolkit
Imports Microsoft.Kinect.Toolkit.FaceTracking
Public Class FaceTrackingViewer
    Public Shared ReadOnly KinectProperty As DependencyProperty = DependencyProperty.Register("Kinect", GetType(KinectSensor), GetType(FaceTrackingViewer), New PropertyMetadata(Nothing, Sub(o, args)
                                                                                                                                                                                              DirectCast(o, FaceTrackingViewer).OnSensorChanged(DirectCast(args.OldValue, KinectSensor), DirectCast(args.NewValue, KinectSensor))
                                                                                                                                                                                          End Sub))

    Const MaxMissedFrames As UInteger = 100
    Dim trackedSkeletons As New Dictionary(Of Integer, SkeletonFaceTracker)()
    Dim myColorImage As Byte()
    Dim myColorImageFormat As ColorImageFormat = ColorImageFormat.Undefined
    Dim myDepthImage As Short()
    Dim myDepthImageFormat As DepthImageFormat = depthImageFormat.Undefined
    Dim myDisposed As Boolean

    Dim skeletonData As Skeleton()

    Private Class SkeletonFaceTracker

        Shared faceTriangles As FaceTriangle()
        Dim facePoints As EnumIndexableCollection(Of FeaturePoint, PointF)
        Dim myFaceTracker As FaceTracker
        Dim lastFaceTrackSucceeded As Boolean
        Dim mySkeletonTrackingState As SkeletonTrackingState

        Property LastTrackedFrame As Integer
        Public Sub Dispose()
            If myFaceTracker Is Nothing = False Then
                myFaceTracker.Dispose()
                myFaceTracker = Nothing
            End If
        End Sub

        Public Sub DrawFaceModel(drawingContext As DrawingContext)
            If lastFaceTrackSucceeded = False OrElse mySkeletonTrackingState <> SkeletonTrackingState.Tracked Then
                Return
            End If

            Dim faceModelPts = New List(Of System.Windows.Point)()
            Dim faceModel = New List(Of FaceModelTriangle)()

            For i As Integer = 0 To facePoints.Count - 1
                faceModelPts.Add(New System.Windows.Point(CInt(Me.facePoints(i).X), CInt(Me.facePoints(i).Y)))
            Next

            For Each t In faceTriangles
                Dim triangle = New FaceModelTriangle()
                triangle.P1 = faceModelPts(t.First)
                triangle.P2 = faceModelPts(t.Second)
                triangle.P3 = faceModelPts(t.Third)
                faceModel.Add(triangle)
            Next

            Dim faceModelGroup = New GeometryGroup()
            For i As Integer = 0 To faceModel.Count - 1
                Dim faceTriangle = New GeometryGroup()
                faceTriangle.Children.Add(New LineGeometry(faceModel(i).P1, faceModel(i).P2))
                faceTriangle.Children.Add(New LineGeometry(faceModel(i).P2, faceModel(i).P3))
                faceTriangle.Children.Add(New LineGeometry(faceModel(i).P3, faceModel(i).P1))
                faceModelGroup.Children.Add(faceTriangle)
            Next

            drawingContext.DrawGeometry(Brushes.Red, New Pen(Brushes.Red, 2.0), faceModelGroup)
        End Sub

        Friend Sub OnFrameReady(kinect As KinectSensor, myColorImageFormat As ColorImageFormat, myColorImage As Byte(), myDepthImageFormat As DepthImageFormat, myDepthImage As Short(), mySkeleton As Skeleton)
            mySkeletonTrackingState = mySkeleton.TrackingState

            If mySkeletonTrackingState <> SkeletonTrackingState.Tracked Then
                Return
            End If

            If myFaceTracker Is Nothing = True Then
                Try
                    myFaceTracker = New FaceTracker(kinect)
                Catch generatedExceptionName As InvalidOperationException
                    myFaceTracker = Nothing
                End Try
            End If

            If myFaceTracker Is Nothing = False Then
                Dim frame As FaceTrackFrame = myFaceTracker.Track(myColorImageFormat, myColorImage, myDepthImageFormat, myDepthImage, mySkeleton)

                lastFaceTrackSucceeded = frame.TrackSuccessful
                If lastFaceTrackSucceeded = True Then
                    If faceTriangles Is Nothing = True Then
                        faceTriangles = frame.GetTriangles()
                    End If

                    facePoints = frame.GetProjected3DShape()
                End If
            End If
        End Sub

        Private Structure FaceModelTriangle
            Public P1 As System.Windows.Point
            Public P2 As System.Windows.Point
            Public P3 As System.Windows.Point
        End Structure
    End Class

    Protected Overrides Sub Finalize()
        Try
            Dispose(False)
        Finally
            MyBase.Finalize()
        End Try
    End Sub

    Public Property Kinect() As KinectSensor
        Get
            Return DirectCast(Me.GetValue(KinectProperty), KinectSensor)
        End Get

        Set(value As KinectSensor)
            Me.SetValue(KinectProperty, value)
        End Set
    End Property

    Public Sub Dispose()
        Me.Dispose(True)
        GC.SuppressFinalize(Me)
    End Sub

    Protected Overridable Sub Dispose(disposing As Boolean)
        If myDisposed = False Then
            myDisposed = True
        End If
    End Sub

    Protected Overrides Sub OnRender(drawingContext As DrawingContext)
        MyBase.OnRender(drawingContext)
        For Each faceInformation As SkeletonFaceTracker In trackedSkeletons.Values
            faceInformation.DrawFaceModel(drawingContext)
        Next
    End Sub

    Private Sub OnAllFramesReady(sender As Object, e As AllFramesReadyEventArgs)
        Dim myColorImageFrame As ColorImageFrame = Nothing
        Dim myDepthImageFrame As DepthImageFrame = Nothing
        Dim mySkeletonFrame As SkeletonFrame = Nothing

        Try
            myColorImageFrame = e.OpenColorImageFrame()
            myDepthImageFrame = e.OpenDepthImageFrame()
            mySkeletonFrame = e.OpenSkeletonFrame()

            If myColorImageFrame Is Nothing OrElse myDepthImageFrame Is Nothing OrElse mySkeletonFrame Is Nothing Then
                Return
            End If

            If myDepthImageFormat <> myDepthImageFrame.Format Then
                myDepthImage = Nothing
                myDepthImageFormat = myDepthImageFrame.Format
            End If

            If myColorImageFormat <> myColorImageFrame.Format Then
                myColorImage = Nothing
                myColorImageFormat = myColorImageFrame.Format
            End If

            If myDepthImage Is Nothing Then
                myDepthImage = New Short(myDepthImageFrame.PixelDataLength - 1) {}
            End If

            If myColorImage Is Nothing Then
                myColorImage = New Byte(myColorImageFrame.PixelDataLength - 1) {}
            End If

            If skeletonData Is Nothing OrElse skeletonData.Length <> mySkeletonFrame.SkeletonArrayLength Then
                skeletonData = New Skeleton(mySkeletonFrame.SkeletonArrayLength - 1) {}
            End If

            myColorImageFrame.CopyPixelDataTo(myColorImage)
            myDepthImageFrame.CopyPixelDataTo(myDepthImage)
            mySkeletonFrame.CopySkeletonDataTo(skeletonData)

            For Each mySkeleton As Skeleton In skeletonData
                If mySkeleton.TrackingState = SkeletonTrackingState.Tracked OrElse mySkeleton.TrackingState = SkeletonTrackingState.PositionOnly Then
                    If trackedSkeletons.ContainsKey(mySkeleton.TrackingId) = False Then
                        trackedSkeletons.Add(mySkeleton.TrackingId, New SkeletonFaceTracker())
                    End If

                    Dim mySkeletonFaceTracker As New SkeletonFaceTracker
                    If trackedSkeletons.TryGetValue(mySkeleton.TrackingId, mySkeletonFaceTracker) = True Then
                        mySkeletonFaceTracker.OnFrameReady(Kinect, myColorImageFormat, myColorImage, myDepthImageFormat, myDepthImage, mySkeleton)
                        mySkeletonFaceTracker.LastTrackedFrame = mySkeletonFrame.FrameNumber
                    End If
                End If
            Next

            InvalidateVisual()
        Finally
            If myColorImageFrame Is Nothing = False Then
                myColorImageFrame.Dispose()
            End If

            If myDepthImageFrame Is Nothing = False Then
                myDepthImageFrame.Dispose()
            End If

            If mySkeletonFrame Is Nothing = False Then
                mySkeletonFrame.Dispose()
            End If
        End Try
    End Sub

    Private Sub OnSensorChanged(oldSensor As KinectSensor, newSensor As KinectSensor)
        If oldSensor Is Nothing = False Then
            RemoveHandler oldSensor.AllFramesReady, AddressOf OnAllFramesReady
        End If

        If newSensor Is Nothing = False Then
            AddHandler newSensor.AllFramesReady, AddressOf OnAllFramesReady
        End If
    End Sub
End Class
